from __future__ import annotations

import os
import subprocess
import sys
import time
from dataclasses import dataclass
from threading import Timer
from typing import TypeVar

import cptk.constants
import cptk.utils
from cptk.core.system import System
from cptk.local.problem import LocalProblem
from cptk.scrape import Test

T = TypeVar('T')


@dataclass
class RunnerResult:
    runner: Runner
    code: int
    timed_out: bool
    outs: str | None = None
    errs: str | None = None


class Nothing:
    def __getattribute__(self: T, *_) -> T:
        return self

    def __call__(self: T, *_, **__) -> T:
        return self


class BakingError(cptk.utils.cptkException):
    def __init__(self, code: int, cmd: str) -> None:
        self.code = code
        self.cmd = cmd
        super().__init__(
            f'Execution of command resulted in exit code {code}:\n{cmd}',
        )


class NoTestConfigurationError(cptk.utils.cptkException):
    def __init__(self) -> None:
        super().__init__("Testing workflow isn't configured for the problem")


class Runner:
    def __init__(self, env: dict = None) -> None:
        self.env = env if env is not None else os.environ

    def exec(
        self,
        cmd: str,
        input: str = None,
        timeout: float = None,
        redirect: bool = True,
        wd: str = None,
    ) -> RunnerResult:
        """ Executes the given command, and returns a 'RunnerResult' instance
        that describes the result of the execution.
        If input is provided, it is piped into the input of the subprocess.
        If timeout is provided, the execution of the process will get
        terminated after the provided amount of seconds. """

        proc = subprocess.Popen(
            cmd.split(),
            cwd=wd,
            env=self.env,
            encoding='utf8',
            stdin=subprocess.PIPE if redirect else sys.stdin,
            stdout=subprocess.PIPE if redirect else sys.stdout,
            stderr=subprocess.PIPE if redirect else sys.stderr,
        )

        # I used a timer thread and a Popen instance instead of the builtin
        # run or communicate methods with the 'timeout' argument to allow
        # capturing the stdin and stderr even if we terminate the process.

        timer = Timer(timeout, proc.kill) if timeout is not None else Nothing()

        if input is not None:
            proc.stdin.write(input)
            proc.stdin.flush()

        try:
            timer.start()
            proc.wait()
        finally:
            timed_out = False if timeout is None else not timer.is_alive()
            timer.cancel()

        return RunnerResult(
            runner=self,
            outs=proc.stdout.read() if redirect else None,
            errs=proc.stderr.read() if redirect else None,
            code=proc.returncode,
            timed_out=timed_out,
        )


class Chef:
    """ Bake, serve and test local problems. """

    def __init__(self, problem: LocalProblem) -> None:
        self._problem = problem
        self._runner = Runner()

    @property
    def _using_string(self) -> str:
        name = self._problem.recipe.name
        if name is not None:
            return f'using recipe {name!r}'
        return 'using the default recipe'

    def bake(self) -> None:
        """ Bakes (generates) the executable of the current problem solution.
        If the recipe configuration file of the current problem doesn't specify
        a 'bake' option, returns None quietly. """

        if not self._problem.recipe.bake:
            return

        location = self._problem.location
        System.log(f'Baking has begun ({self._using_string})')
        start = time.time()

        for cmd in self._problem.recipe.bake:
            System.details(cmd)
            res = self._runner.exec(cmd, wd=location, redirect=False)
            if res.code:
                raise BakingError(res.code, cmd)

        seconds = time.time() - start
        System.log(f'Solution is baked! (took {seconds:.02f} seconds)')

    def serve(self) -> None:
        """ Bakes the local problem (if a baking recipe is provided), and serves
        it while piping the standard input to the executable. """

        self.bake()

        cmd = self._problem.recipe.serve
        location = self._problem.location

        System.log(f'Serving solution ({self._using_string})')
        System.details(cmd)
        res = self._runner.exec(cmd, wd=location, redirect=False)
        System.abort(res.code)

    def _load_tests(self) -> dict[str, Test]:
        """ Returns a list of tests where the keys are their names and the
        values are the test details. """

        # TODO: the ideal solution will provide a way for the user to fully
        # configure the way that the test inputs end expectations are stored.
        # For now, we force a standard that uses treats all .in files inside
        # the folder as input files, and all .out files as the expected outputs

        folder = os.path.join(
            self._problem.location,
            self._problem.recipe.test.folder,
        )
        if not os.path.isdir(folder):
            return dict()
        files = [
            item for item in os.listdir(
                folder,
            ) if os.path.isfile(os.path.join(folder, item))
        ]

        inputs = {
            filename[:-len(cptk.constants.INPUT_FILE_SUFFIX)]
            for filename in files
            if filename.endswith(cptk.constants.INPUT_FILE_SUFFIX)
        }

        outputs = {
            filename[:-len(cptk.constants.OUTPUT_FILE_SUFFIX)]
            for filename in files
            if filename.endswith(cptk.constants.OUTPUT_FILE_SUFFIX)
        }

        res = dict()
        for name in inputs.intersection(outputs):
            inpp = os.path.join(folder, name + cptk.constants.INPUT_FILE_SUFFIX)
            with open(inpp, 'r', encoding='utf8') as file:
                inp = file.read()
            outp = os.path.join(folder, name + cptk.constants.OUTPUT_FILE_SUFFIX)
            with open(outp, 'r', encoding='utf8') as file:
                out = file.read()
            res[name] = Test(inp, out)

        return res

    def test(self) -> None:
        """ Bakes (if a baking recipe is provided) and serves the local tests
        that are linked to the problem. """

        if self._problem.recipe.test is None:
            raise NoTestConfigurationError()

        self.bake()
        cmd = self._problem.recipe.serve
        location = self._problem.location
        timeout = self._problem.recipe.test.timeout

        tests = self._load_tests()

        LogFunc = System.title if tests else System.warn
        LogFunc(f'Found {len(tests)} tests')

        passed = 0
        start = time.time()

        for name, test in sorted(tests.items()):
            res = self._runner.exec(
                cmd, wd=location, input=test.input,
                redirect=True, timeout=timeout,
            )

            if res.timed_out:
                System.error('Execution timed out', title=name)
            elif res.code:
                System.error(f'Nonzero exit code {res.code}', title=name)
            elif test.expected is not None and res.outs != test.expected:
                System.error('Output differs from expectation', title=name)
            else:
                System.success('Output matches expectations', title=name)
                passed += 1

        seconds = time.time() - start
        failed = len(tests) - passed
        System.title(
            f'{passed} passed and {failed} failed'
            f' in {seconds:.2f} seconds',
        )

        System.abort(1 if failed else 0)



from __future__ import annotations

import os
from dataclasses import dataclass


@dataclass
class Template:
    uid: str
    name: str
    path: str


HERE = os.path.dirname(__file__)
ROOT = os.path.join(HERE, '..')
TEMPLATES_FOLDER = os.path.normpath(os.path.join(ROOT, 'defaults', 'templates'))

DEFAULT_TEMPLATES = (
    Template(
        uid='g++',
        name='C++ (Compiled with g++)',
        path=os.path.join(TEMPLATES_FOLDER, 'g++'),
    ),
    Template(
        uid='py',
        name='Python 3',
        path=os.path.join(TEMPLATES_FOLDER, 'py'),
    ),
)


__all__ = ['Template', 'DEFAULT_TEMPLATES']



__version__ = '0.1.0a2'
__author__ = 'Alon Krymgand Osovsky'
__author_email__ = 'downtown2u@gmail.com'
__description__ = ' '.join((
    'Your personal assistant for everything',
    'competitive-programming-related.',
))



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from urllib import parse

from cptk.scrape import Contest
from cptk.scrape import Problem
from cptk.scrape import Scraped
from cptk.scrape import Test
from cptk.scrape import Website

if TYPE_CHECKING:
    from cptk.scrape import PageInfo
    from bs4 import BeautifulSoup


@dataclass(unsafe_hash=True)
class CodeforecsProblem(Problem):
    mark: str = field(compare=False, default=None)
    gym: bool = field(compare=False, default=False)


@dataclass(unsafe_hash=True)
class CodeforcesContest(Contest):
    pass


@dataclass(unsafe_hash=True)
class CodeforecsGroup(Scraped):
    pass


@dataclass(unsafe_hash=True)
class CodeforcesGroupProblem(CodeforecsProblem):
    """ A problem that is a part of a contest that is organized by a Codeforecs
    group, and is not official. """

    group: CodeforecsGroup = None


class Codeforces(Website):

    @property
    def name(self) -> str:
        return 'Codeforces'

    @property
    def domain(self) -> str:
        return 'codeforces.com'

    @staticmethod
    def _parse_code_text(soup: BeautifulSoup) -> None:
        text = soup.text.replace('<br>', '\n').replace('\r', '').strip()
        return '\n'.join(
            line.strip()
            for line in text.splitlines(keepends=False)
        ) + '\n'

    @classmethod
    def _parse_tests(cls, info: PageInfo) -> list[Test]:
        """ Assumes that the given 'PageInfo' instance describes a Problem page,
        and parses all sample test cases in the page. """

        tests_soup = info.data.find('div', {'class': 'sample-tests'})
        for br_soup in tests_soup.find_all('br'):
            br_soup.replace_with('\n')

        inputs_soup = tests_soup.find_all('div', {'class': 'input'})
        outputs_soup = tests_soup.find_all('div', {'class': 'output'})

        return [
            Test(
                input=cls._parse_code_text(in_soup.find('pre')),
                expected=cls._parse_code_text(ex_soup.find('pre')),
            ) for in_soup, ex_soup in zip(inputs_soup, outputs_soup)
        ]

    def is_problem(self, info: PageInfo) -> bool:
        """ Returns 'True' if the given 'PageInfo' instance contains a problem
        statement. """
        elem = info.data.find('div', {'class': 'problem-statement'})
        return elem is not None

    def to_problem(self, info: PageInfo) -> CodeforecsProblem | None:
        """ Assumes that the given 'PageInfo' instance contains a problem
        statement and returns a 'Problem' instance that describes the problem.
        """

        header_soup = info.data.find('div', {'class': 'header'})

        title = header_soup.find('div', {'class': 'title'}).text
        mark, name = (i.strip() for i in title.split('.'))

        contest = self._contest_from_sidebar(info)

        gym = next(
            p.lower() == 'gym'
            for p in parse.urlparse(info.url).path.split('/')
            if p
        )

        time_limit_soup = header_soup.find('div', {'class': 'time-limit'})
        time_limit = next(
            float(word) for word
            in time_limit_soup.find(text=True, recursive=False).split()
            if word.strip().isnumeric()
        )

        memory_limit_soup = header_soup.find('div', {'class': 'memory-limit'})
        memory_limit = next(
            float(word) for word
            in memory_limit_soup.find(text=True, recursive=False).split()
            if word.strip().isnumeric()
        )

        group = self._group_from_sidebar(info)

        kwargs = {
            '_uid': [contest._uid, mark],
            'website': self,
            'name': name,
            'mark': mark,
            'gym': gym,
            'url': info.url,
            'contest': contest,
            'tests': self._parse_tests(info),
            'time_limit': time_limit,
            'memory_limit': memory_limit,
        }

        if group:
            return CodeforcesGroupProblem(**kwargs, group=group)

        return CodeforecsProblem(**kwargs)

    def _group_from_sidebar(
        self,
        info: PageInfo,
    ) -> CodeforecsGroup | None:

        tables = info.data.find_all('table', {'class': 'rtable'})
        links = [table.find('a', href=True) for table in tables]

        try:
            # Searching for a that looks like: '/group/abc123/'
            link = next(
                link
                for link in links
                if link is not None
                if link['href'].count('/') == 2
                if link['href'].split('/')[1] == 'group'
            )

        except StopIteration:
            return None

        return CodeforecsGroup(
            _uid=link['href'].split('/')[-1],
            website=self,
            name=link.text.strip(),
            url=parse.urljoin(info.url, link['href']),
        )

    def _contest_from_sidebar(
        self,
        info: PageInfo,
    ) -> CodeforcesContest | None:
        """ Tries to pull information about the current contest using the
        sidebar information that is displayed on every page that is related to
        a contest. If fails to locate the sidebar, returns None. """

        tables = info.data.find_all('table', {'class': 'rtable'})
        links = [table.find('a', href=True) for table in tables]

        try:
            link = next(
                link for link in links
                if not {'contest', 'gym'}.isdisjoint(link['href'].split('/'))
            )
        except StopIteration:
            return None

        return CodeforcesContest(
            website=self,
            _uid=int(link['href'].split('/')[-1]),
            name=link.text.strip(),
            url=parse.urljoin(info.url, link['href']),
        )



from __future__ import annotations

import re
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from urllib.parse import urljoin
from urllib.parse import urlparse

from cptk.scrape import Contest
from cptk.scrape import Problem
from cptk.scrape import Test
from cptk.scrape import Website

if TYPE_CHECKING:
    from cptk import PageInfo
    from bs4 import BeautifulSoup


@dataclass(unsafe_hash=True)
class CsesContest(Contest):
    pass


@dataclass(unsafe_hash=True)
class CsesContestProblem(Problem):
    mark: str = field(compare=False, default=None)


@dataclass(unsafe_hash=True)
class CsesProblemsetProblem(Problem):
    section: str = field(compare=False, default=None)


class Cses(Website):

    @property
    def name(self) -> str:
        return 'Code Submission Evaluation System'

    @property
    def domain(self) -> str:
        return 'cses.fi'

    def _contest_from_titlebar(
        self,
        info: PageInfo,
    ) -> CsesContest | None:
        titlebar_soup = info.data.find('div', {'class': 'title-block'})

        url = urlparse(info.url)
        try:
            uid = next(
                p for p in url.path.split('/')
                if p.strip()
            )
        except StopIteration:
            return None

        try:
            title_soup = titlebar_soup.find('h3')
            url = urljoin(info.url, title_soup.find('a')['href'])

            return CsesContest(
                website=self,
                _uid=uid,
                name=title_soup.text.strip(),
                url=url,
            )

        except (AttributeError, TypeError):
            return None

    @staticmethod
    def _parse_code_text(soup: BeautifulSoup) -> None:
        return soup.text.replace('<br>', '\n').replace('\r', '').strip() + '\n'

    def _parse_tests(self, info: PageInfo) -> list[Test]:
        content = info.data.find('div', {'class': 'content'})

        titles_soup = content.find_all(
            'b', {'id': re.compile('.*(example|esimerkki).*')},
        )

        tests = list()
        for title in titles_soup:
            title: BeautifulSoup

            input_soup = title.find_next_sibling('code')
            expected_soup = input_soup.find_next_sibling('code')

            tests.append(
                Test(
                    input=self._parse_code_text(input_soup),
                    expected=self._parse_code_text(expected_soup),
                ),
            )

        return tests

    def is_problem(self, info: PageInfo) -> bool:
        soup = info.data.find('ul', {'class': 'task-constraints'})
        return soup is not None

    def to_problem(self, info: PageInfo) -> Problem | None:
        contest = self._contest_from_titlebar(info)

        title_soup = info.data.find('div', {'class': 'title-block'})
        content_soup = info.data.find('div', {'class': 'content'})
        sidebar_soup = info.data.find('div', {'class': 'nav sidebar'})
        current_soup = sidebar_soup.find('a', {'class': 'current'})
        constrains_soup = content_soup.find(
            'ul', {'class': 'task-constraints'},
        )

        uid_soup = content_soup.find('input', {'name': 'task'})
        uid = int(uid_soup['value'])

        level_soup = current_soup.find('b')
        level = level_soup.text.strip() if level_soup is not None else None

        name = title_soup.find('h1').text.strip()

        section_soup = sidebar_soup.find('h4')
        section = section_soup.text.strip()

        time_limit_soup = constrains_soup.find('li')
        time_limit = next(
            float(word) for word
            in time_limit_soup.find(text=True, recursive=False)
            if word.strip().isnumeric()
        )

        memory_limit_soup = time_limit_soup.find_next_sibling('li')
        memory_limit = next(
            float(word) for word
            in memory_limit_soup.find(text=True, recursive=False).split()
            if word.strip().isnumeric()
        )

        kwargs = {
            'website': self,
            '_uid': [contest._uid, uid],
            'name': name,
            'url': info.url,
            'tests': self._parse_tests(info),
            'contest': contest,
            'time_limit': time_limit,
            'memory_limit': memory_limit,
        }

        if level is not None:
            return CsesContestProblem(**kwargs, mark=level)
        else:
            return CsesProblemsetProblem(**kwargs, section=section)



from __future__ import annotations

RECIPE_FILE = 'recipes.cptk.yaml'
PROJECT_FILE = '.cptk/project.cptk.yaml'
LAST_FILE = '.cptk/stayaway/last.cptk.txt'

MOVE_FILE = '.cptk/moves.cptk.txt'
MOVE_FILE_SEPERATOR = '::'
MOVE_SAFES = ['./', '**/.cptk/**']

LAST_FILE_SEPERATOR = '::'

INPUT_FILE_SUFFIX = '.in'
OUTPUT_FILE_SUFFIX = '.out'


def TEST_NAME_GENERATOR():
    n = 1
    while True:
        yield f'sample{n:02d}'
        n += 1



from __future__ import annotations

from cptk.core.chef import BakingError
from cptk.core.chef import NoTestConfigurationError
from cptk.core.config import ConfigFileError
from cptk.core.config import ConfigFileNotFound
from cptk.core.config import ConfigFileParsingError
from cptk.core.config import ConfigFileValueError
from cptk.core.fetcher import InvalidClone
from cptk.core.fetcher import UnknownWebsite
from cptk.core.preprocessor import PreprocessError
from cptk.core.preprocessor import PreprocessFileError
from cptk.core.preprocessor import PreprocessNameError
from cptk.core.preprocessor import PreprocessStringError
from cptk.core.system import SystemRunError
from cptk.local.problem import NoRecipesFound
from cptk.local.problem import RecipeNameNotFound
from cptk.local.problem import RecipeNotFoundError
from cptk.local.project import InvalidMoveDest
from cptk.local.project import InvalidMovePath
from cptk.local.project import InvalidMoveSource
from cptk.local.project import ProjectNotFound
from cptk.utils import cptkException

__all__ = [
    # cptk.core.chef
    'BakingError',
    'NoTestConfigurationError',

    # cptk.core.config
    'ConfigFileError',
    'ConfigFileNotFound',
    'ConfigFileParsingError',
    'ConfigFileValueError',

    # cptk.core.integrator
    'InvalidClone',
    'UnknownWebsite',

    # cptk.core.preprocessor
    'PreprocessError',
    'PreprocessFileError',
    'PreprocessNameError',
    'PreprocessStringError',

    # cptk.core.system
    'SystemRunError',

    # cptk.local.problem
    'NoRecipesFound',
    'RecipeNameNotFound',
    'RecipeNotFoundError',

    # cptk.local.project
    'InvalidMoveDest',
    'InvalidMovePath',
    'InvalidMoveSource',
    'ProjectNotFound',

    # cptk.utils
    'cptkException',
]



from __future__ import annotations

from setuptools import find_packages
from setuptools import setup

import cptk

with open("README.md", encoding="utf8") as f:
    README = f.read()

with open("CHANGELOG.md", encoding="utf8") as f:
    CHANGELOG = f.read()

with open("requirements.txt", encoding="utf8") as f:
    DEPENDENCIES = f.read().splitlines()

setup(
    name="cptk",
    version=cptk.__version__,
    description=cptk.__description__,
    url='https://github.com/RealA10N/cptk',

    python_requires=">=3.7,<4",
    install_requires=DEPENDENCIES,

    long_description=README + '\n\n' + CHANGELOG,
    long_description_content_type="text/markdown",

    author=cptk.__author__,
    author_email=cptk.__author_email__,
    packages=find_packages(include=['cptk*']),

    entry_points={
        "cptk_sites": [
            "codeforces=cptk.websites:Codeforces",
            "csesfi=cptk.websites:Cses",
            "kattis=cptk.websites:Kattis",
        ],
        "console_scripts": [
            "cptk=cptk.main:main",
        ],
    },
)



from __future__ import annotations

from colorama import deinit
from colorama import init

import cptk.commands
from cptk.core.system import System
from cptk.utils import cptkException


def main(args: list[str] | None = None) -> int:

    code = 0  # os.EX_OK

    try:
        init()
        cptk.commands.collector.run(args)

    except SystemExit as err:
        # argparse throws a SystemExit if fails to parse arguments or
        # if some arguments are invalid. The System.abort function throws an
        # SystemExit exception too.
        code = err.code

    except cptkException as err:
        System.error(err)
        code = 65  # os.EX_DATAERR

    except Exception as err:
        System.unexpected_error(err)
        code = 70  # os.EX_SOFTWARE

    finally:
        deinit()
        raise SystemExit(code)


if __name__ == '__main__':
    main()



from __future__ import annotations

import os
import re
import shutil
from argparse import ArgumentTypeError
from functools import lru_cache
from typing import Callable
from typing import Generator
from typing import TypeVar
T = TypeVar('T')


class cptkException(Exception):
    """ Base cptk exception. All exceptions raised and created by cptk should
    inherent from this one. Note that this expection can't be defined in
    'exceptions.py' to avoid circular import errors. (: """


def cached_property(f: Callable[..., T]) -> T:
    """ The builtin cached_property decorator was introduced in py3.8.
    This is a simple implementation of the decortor that can be ran with older
    Python versions. https://tinyurl.com/smhlght """

    return property(lru_cache(None)(f))


def find_tree_files(dir: str) -> Generator[str, None, None]:
    """ Yields all files in the directory as absolute paths. """

    if not os.path.isdir(dir):
        return

    for name in os.listdir(dir):
        path = os.path.join(dir, name)
        if os.path.isdir(path):
            yield from find_tree_files(path)
        elif os.path.isfile(path):
            yield path


def find_common_files(a: str, b: str) -> list[str]:
    """ Returns a list of files that share the same relative path in both
    directories a and b. """

    # To preserve the order of the files, we use the order of a_files
    a_files = list(os.path.relpath(p, start=a) for p in find_tree_files(a))
    b_files = (os.path.relpath(p, start=b) for p in find_tree_files(b))
    commons = set(a_files).intersection(b_files)
    return [p for p in a_files if p in commons]


def soft_tree_copy(src: str, dst: str) -> None:
    """ Copies all files from the source directory into the destination
    directory, recursively. If the file already exists in the destination
    directory, it will be overwritten. If some subdirectory src/a exists, it
    will be created in the destintation directory, even if there are no files
    inside it in the source directory. """

    os.makedirs(dst, exist_ok=True)

    for name in os.listdir(src):
        item_src = os.path.join(src, name)
        item_dst = os.path.join(dst, name)

        if os.path.isdir(item_src):
            soft_tree_copy(item_src, item_dst)
        elif os.path.isfile(item_src):
            shutil.copyfile(item_src, item_dst)


@lru_cache(None)
def _url_regex() -> str:
    # Cached so the regex is compiled only once,
    # and only if needed at all.
    return re.compile(
        r"(\w+://)?"                # protocol (optional)
        r"(\w+\.)?"                 # host (optional)
        r"((\w+)\.(\w+))"           # domain
        r"(\.\w+)*"                 # top-level domain (optional, can have > 1)
        r"([\w\-\._\~/]*)*(?<!\.)",  # path, params, anchors, etc. (optional)
    )


def valid_url(url) -> bool:
    return bool(_url_regex().fullmatch(url))


def path_validator(
    dir_ok=True,
    file_ok=True,
    must_exist=False,
) -> Callable[[str], str]:
    def validator(path: str) -> str:
        path = os.path.normpath(path)

        if not file_ok and os.path.isfile(path):
            raise ArgumentTypeError(f"{path!r} is a file")

        if not dir_ok and os.path.isdir(path):
            raise ArgumentTypeError(f"{path!r} is a direcotry")

        if must_exist and not os.path.exists(path):
            raise ArgumentTypeError(f"path {path!r} doesn't exist")

        return path
    return validator


def url_validator(url: str) -> str:
    if not valid_url(url):
        raise ArgumentTypeError(f'invalid url {url!r}')
    return url



from __future__ import annotations

import os
from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

import pydantic

import cptk.constants
import cptk.scrape
import cptk.utils
from cptk.core.config import ConfigFileError
from cptk.core.config import ConfigFileNotFound
from cptk.core.config import Configuration
from cptk.core.system import System

if TYPE_CHECKING:
    from cptk.core.preprocessor import Preprocessor

T = TypeVar('T')


class RecipeNotFoundError(cptk.utils.cptkException):
    def __init__(self, filepath: str, msg: str) -> None:
        self.filepath = filepath
        super().__init__(msg)


class RecipeNameNotFound(RecipeNotFoundError):
    def __init__(self, filepath: str, name: str) -> None:
        self.name = name
        super().__init__(
            filepath,
            f'Recipe named {name!r} not found in recipes file {filepath!r}',
        )


class NoRecipesFound(RecipeNotFoundError):
    def __init__(self, filepath: str) -> None:
        super().__init__(
            filepath,
            f'No recipes found in recipes file {filepath!r}',
        )


class TestRecipe(pydantic.BaseModel):
    folder: str
    timeout: Union[float, str, None] = None

    def preprocess(self: T, processor: Preprocessor) -> type[T]:
        def parse_str(v):
            if isinstance(v, str):
                return processor.parse_string(v)
            else:
                return v

        kwargs = {
            'folder': parse_str(self.folder),
            'timeout': parse_str(self.timeout),
        }

        return type(self)(**kwargs)


class Recipe(pydantic.BaseModel):

    name: Optional[str] = None
    bake: List[str] = []
    serve: str
    test: Optional[TestRecipe] = None

    @pydantic.validator('bake', pre=True)
    @classmethod
    def string_to_commands(cls, val) -> List[str]:
        if isinstance(val, str):
            return val.split('\n')
        return val

    def preprocess(self: T, processor: Preprocessor) -> type[T]:
        def parse_str(v):
            if isinstance(v, str):
                return processor.parse_string(v)
            else:
                return v

        kwargs = {
            'name': parse_str(self.name),
            'bake': [parse_str(v) for v in self.bake],
            'serve': parse_str(self.serve),
            'test': self.test.preprocess(processor) if self.test else None,
        }

        return type(self)(**kwargs)


class RecipesConfig(Configuration):
    recipes: List[Recipe]


@dataclass(unsafe_hash=True)
class LocalProblem:
    location: str = field(compare=True)
    name: str | None = field(compare=True, default=None)

    @staticmethod
    def _store_test(folder: str, name: str, test: cptk.scrape.Test) -> None:

        inp = os.path.join(folder, name + cptk.constants.INPUT_FILE_SUFFIX)
        with open(inp, 'w', encoding='utf8') as file:
            file.write(test.input)

        out = os.path.join(folder, name + cptk.constants.OUTPUT_FILE_SUFFIX)
        if test.expected is not None:
            with open(out, 'w', encoding='utf8') as file:
                file.write(test.expected)

    def store_tests(self, folder: str, tests: list[cptk.scrape.Test]) -> None:
        if not tests:
            return
        gen = cptk.constants.TEST_NAME_GENERATOR()
        folder = os.path.join(self.location, folder)
        os.makedirs(folder, exist_ok=True)
        for test in tests:
            name = next(gen)
            self._store_test(folder, name, test)

    @classmethod
    def init(cls: type[T], location: str, recipe: Recipe) -> T:
        path = os.path.join(location, cptk.constants.RECIPE_FILE)
        config = None

        try:
            config = RecipesConfig.load(path)
        except ConfigFileNotFound:
            pass  # File doesn't exists -> we will create a new one later.
        except ConfigFileError:
            System.warn(
                f'Failed to parse existing recipe file {path!r}. '
                'Continuing the process will overwrite the file.',
            )
            ans = System.confirm('Are you sure you want to continue')
            if not ans:
                System.abort(0)

        if config is None:
            config = RecipesConfig(recipes=list())

        config.recipes.append(recipe)
        config.dump(path)

        return cls(location, recipe.name)

    @staticmethod
    def is_problem(location: str, name: str = None) -> bool:
        """ Returns True if the given directory contains a recipe with the given
        name. If a name isn't provided, returns True if a recipe file is located
        inside the given direcotry and it contains at least one recipe.  """

        path = os.path.join(location, cptk.constants.RECIPE_FILE)

        try:
            recipes = RecipesConfig.load(path)
        except ConfigFileError:
            return False

        names = {r.name for r in recipes.recipes}
        return len(names) > 0 if name is None else name in names

    @property
    def recipe(self) -> Recipe:
        path = os.path.join(self.location, cptk.constants.RECIPE_FILE)
        recipes = RecipesConfig.load(path).recipes

        if not recipes:
            raise NoRecipesFound(path)
        if self.name is None:
            return recipes[0]

        try:
            return next(r for r in recipes if r.name == self.name)
        except StopIteration:
            raise RecipeNameNotFound(path, self.name)
